/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComSlice.h
    \brief    slice header and SPS class (header)
*/

#ifndef __TCOMSLICE__
#define __TCOMSLICE__


#include "CommonDef.h"
#include "TComList.h"

class TComPic;

// ====================================================================================================================
// Class definition
// ====================================================================================================================

/// SPS class
class TComSPS
{
private:
  // Structure
  UInt        m_uiWidth;
  UInt        m_uiHeight;
  Int         m_aiPad[2];
  UInt        m_uiMaxCUWidth;
  UInt        m_uiMaxCUHeight;
  UInt        m_uiMaxCUDepth;
  UInt        m_uiMinTrDepth;
  UInt        m_uiMaxTrDepth;
  
  // Tool list
  UInt        m_uiQuadtreeTULog2MaxSize;
  UInt        m_uiQuadtreeTULog2MinSize;
  UInt        m_uiQuadtreeTUMaxDepthInter;
  UInt        m_uiQuadtreeTUMaxDepthIntra;
  Bool        m_bUseALF;
  Bool        m_bUseDQP;
  Bool        m_bUseLDC;
  Bool        m_bUsePAD;
  Bool        m_bUseMRG; // SOPH:

#if LM_CHROMA 
  Bool        m_bUseLMChroma; // JL:
#endif

#if DCM_COMB_LIST
  Bool        m_bUseLComb;
  Bool        m_bLCMod;
#endif
  
#if HHI_RMP_SWITCH
  Bool        m_bUseRMP;
#endif
  // Parameter
  AMVP_MODE   m_aeAMVPMode[MAX_CU_DEPTH];
  UInt        m_uiBitDepth;
  UInt        m_uiBitIncrement;
  
  // Max physical transform size
  UInt        m_uiMaxTrSize;
  
  Int m_iAMPAcc[MAX_CU_DEPTH];

#if MTK_NONCROSS_INLOOP_FILTER
  Bool        m_bLFCrossSliceBoundaryFlag;
#endif
#if MTK_SAO
  Bool        m_bUseSAO; 
#endif

public:
  TComSPS();
  virtual ~TComSPS();
  
  // structure
  Void setWidth       ( UInt u ) { m_uiWidth = u;           }
  UInt getWidth       ()         { return  m_uiWidth;       }
  Void setHeight      ( UInt u ) { m_uiHeight = u;          }
  UInt getHeight      ()         { return  m_uiHeight;      }
  Void setMaxCUWidth  ( UInt u ) { m_uiMaxCUWidth = u;      }
  UInt getMaxCUWidth  ()         { return  m_uiMaxCUWidth;  }
  Void setMaxCUHeight ( UInt u ) { m_uiMaxCUHeight = u;     }
  UInt getMaxCUHeight ()         { return  m_uiMaxCUHeight; }
  Void setMaxCUDepth  ( UInt u ) { m_uiMaxCUDepth = u;      }
  UInt getMaxCUDepth  ()         { return  m_uiMaxCUDepth;  }
  Void setMinTrDepth  ( UInt u ) { m_uiMinTrDepth = u;      }
  UInt getMinTrDepth  ()         { return  m_uiMinTrDepth;  }
  Void setMaxTrDepth  ( UInt u ) { m_uiMaxTrDepth = u;      }
  UInt getMaxTrDepth  ()         { return  m_uiMaxTrDepth;  }
  Void setQuadtreeTULog2MaxSize( UInt u ) { m_uiQuadtreeTULog2MaxSize = u;    }
  UInt getQuadtreeTULog2MaxSize()         { return m_uiQuadtreeTULog2MaxSize; }
  Void setQuadtreeTULog2MinSize( UInt u ) { m_uiQuadtreeTULog2MinSize = u;    }
  UInt getQuadtreeTULog2MinSize()         { return m_uiQuadtreeTULog2MinSize; }
  Void setQuadtreeTUMaxDepthInter( UInt u ) { m_uiQuadtreeTUMaxDepthInter = u;    }
  Void setQuadtreeTUMaxDepthIntra( UInt u ) { m_uiQuadtreeTUMaxDepthIntra = u;    }
  UInt getQuadtreeTUMaxDepthInter()         { return m_uiQuadtreeTUMaxDepthInter; }
  UInt getQuadtreeTUMaxDepthIntra()         { return m_uiQuadtreeTUMaxDepthIntra; }
  Void setPad         (Int iPad[2]) { m_aiPad[0] = iPad[0]; m_aiPad[1] = iPad[1]; }
  Void setPadX        ( Int  u ) { m_aiPad[0] = u; }
  Void setPadY        ( Int  u ) { m_aiPad[1] = u; }
  Int  getPad         ( Int  u ) { assert(u < 2); return m_aiPad[u];}
  Int* getPad         ( )        { return m_aiPad; }
  
  // physical transform
  Void setMaxTrSize   ( UInt u ) { m_uiMaxTrSize = u;       }
  UInt getMaxTrSize   ()         { return  m_uiMaxTrSize;   }
  
  // Tool list
  Bool getUseALF      ()         { return m_bUseALF;        }
  Bool getUseDQP      ()         { return m_bUseDQP;        }
  
  Bool getUseLDC      ()         { return m_bUseLDC;        }
  Bool getUsePAD      ()         { return m_bUsePAD;        }
  Bool getUseMRG      ()         { return m_bUseMRG;        } // SOPH:
  
  Void setUseALF      ( Bool b ) { m_bUseALF  = b;          }
  Void setUseDQP      ( Bool b ) { m_bUseDQP   = b;         }
  
  Void setUseLDC      ( Bool b ) { m_bUseLDC   = b;         }
  Void setUsePAD      ( Bool b ) { m_bUsePAD   = b;         }
  Void setUseMRG      ( Bool b ) { m_bUseMRG  = b;          } // SOPH:
  
#if DCM_COMB_LIST
  Void setUseLComb    (Bool b)   { m_bUseLComb = b;         }
  Bool getUseLComb    ()         { return m_bUseLComb;      }
  Void setLCMod       (Bool b)   { m_bLCMod = b;     }
  Bool getLCMod       ()         { return m_bLCMod;  }
#endif

#if HHI_RMP_SWITCH
  Bool getUseRMP     ()         { return m_bUseRMP; }
  Void setUseRMP     ( Bool b ) { m_bUseRMP = b;    }
#endif
  
#if LM_CHROMA 
  Bool getUseLMChroma ()         { return m_bUseLMChroma;        }
  Void setUseLMChroma ( Bool b ) { m_bUseLMChroma  = b;          }
#endif

  // AMVP mode (for each depth)
  AMVP_MODE getAMVPMode ( UInt uiDepth ) { assert(uiDepth < g_uiMaxCUDepth);  return m_aeAMVPMode[uiDepth]; }
  Void      setAMVPMode ( UInt uiDepth, AMVP_MODE eMode) { assert(uiDepth < g_uiMaxCUDepth);  m_aeAMVPMode[uiDepth] = eMode; }
  
  // Bit-depth
  UInt      getBitDepth     ()         { return m_uiBitDepth;     }
  Void      setBitDepth     ( UInt u ) { m_uiBitDepth = u;        }
  UInt      getBitIncrement ()         { return m_uiBitIncrement; }
  Void      setBitIncrement ( UInt u ) { m_uiBitIncrement = u;    }

#if MTK_NONCROSS_INLOOP_FILTER
  Void      setLFCrossSliceBoundaryFlag     ( Bool   bValue  )    { m_bLFCrossSliceBoundaryFlag = bValue; }
  Bool      getLFCrossSliceBoundaryFlag     ()                    { return m_bLFCrossSliceBoundaryFlag;   } 
#endif

#if MTK_SAO
  Void setUseSAO                  (Bool bVal)  {m_bUseSAO = bVal;}
  Bool getUseSAO                  ()           {return m_bUseSAO;}
#endif

};

/// PPS class
class TComPPS
{
private:
#if CONSTRAINED_INTRA_PRED
  Bool        m_bConstrainedIntraPred;    //  constrained_intra_pred_flag
#endif
  
public:
  TComPPS();
  virtual ~TComPPS();
  
#if CONSTRAINED_INTRA_PRED
  Bool      getConstrainedIntraPred ()         { return  m_bConstrainedIntraPred; }
  Void      setConstrainedIntraPred ( Bool b ) { m_bConstrainedIntraPred = b;     }
#endif
};
#if SONY_QM
typedef struct
{
  Bool bIsDpcm;           ///< flag for using DPCM
  Bool bIsRun;            ///< flag for using Run length
  Bool bIsSign;           ///< flag for using VLC
  Bool bIsSymmetry;       ///< flag for using Symmetry
  Bool bIsResidual;       ///< flag for using Residual coding
  UInt uiTableNumRun;     ///< VLC table number for run
  UInt uiTableNumData;    ///< VLC table number for data
  UInt uiPredMode;        ///< prediction mode
  UInt uiPredQMPSIdx;     ///< QMPS index
  Int iPredSizeIdx;       ///< size index for prediction
  Int iPredListIdx;       ///< list index for prediction
} estQMStruct;

/// QM class
class TComQM
{
private:
  UInt     m_uiqmps_id;                                    ///< QMPS index
  Bool     m_use_default_only_flag;                        ///< flag for using default matrix
  UInt     m_pred_mode     [QM_SIZE_NUM][QM_LIST_NUM];     ///< pridction mode
  UInt     m_pred_qmps_id  [QM_SIZE_NUM][QM_LIST_NUM];     ///< reference QMPS index
  UInt     m_pred_size_idc [QM_SIZE_NUM][QM_LIST_NUM];     ///< reference size index
  UInt     m_pred_matrix_id[QM_SIZE_NUM][QM_LIST_NUM];     ///< reference list index
  Bool     m_residual_flag [QM_SIZE_NUM][QM_LIST_NUM];     ///< flag for using residual coding
  Bool     m_dpcm_flag     [QM_SIZE_NUM][QM_LIST_NUM];     ///< flag for using DPCM
  Bool     m_run_flag      [QM_SIZE_NUM][QM_LIST_NUM];     ///< flag for using Run length
  Bool     m_symmetric_flag[QM_SIZE_NUM][QM_LIST_NUM];     ///< flag for using Symmetric mode
  Int      *m_QM4x4      [QM_MAX_QMPS_SIZE][QM_LIST_NUM];  ///< quantization matrix 4x4
  Int      *m_QM8x8      [QM_MAX_QMPS_SIZE][QM_LIST_NUM];  ///< quantization matrix 8x8
  Int      *m_QM16x16    [QM_MAX_QMPS_SIZE][QM_LIST_NUM];  ///< quantization matrix 16x16
  Int      *m_QM32x32    [QM_MAX_QMPS_SIZE][QM_LIST_NUM];  ///< quantization matrix 32x32
  Int      *m_QM4x4_Org  [QM_MAX_QMPS_SIZE][QM_LIST_NUM];  ///< default quantization matrix 4x4
  Int      *m_QM8x8_Org  [QM_MAX_QMPS_SIZE][QM_LIST_NUM];  ///< default quantization matrix 8x8
  Int      *m_QM16x16_Org[QM_MAX_QMPS_SIZE][QM_LIST_NUM];  ///< default quantization matrix 16x16
  Int      *m_QM32x32_Org[QM_MAX_QMPS_SIZE][QM_LIST_NUM];  ///< default quantization matrix 32x32
  Int      m_Qscale[4];                                    ///< qscale for 4 area of quantization matrix
  UInt     m_vlc_table_num_run[QM_SIZE_NUM][QM_LIST_NUM];  ///< VLC table number for run
  UInt     m_vlc_table_num_data[QM_SIZE_NUM][QM_LIST_NUM]; ///< VLC table number for data
  Bool     m_sign_flag [QM_SIZE_NUM][QM_LIST_NUM];         ///< flag for using UVLC
#if TI_QM
  Int      m_QMCOption[4];                                  ///< QMtx compression options for 4x4, 8x8, 16x16 and 32x32 
#endif

public:
  TComQM();
  virtual ~TComQM();
  Void     init                        (Void);
  Void     destroy                     (Void);
  Void     set_qmps_id                 (UInt u)                                      { m_uiqmps_id = u;            }
  UInt     get_qmps_id                 ()                                            { return m_uiqmps_id;         }
  Void     set_use_default_only_flag   (Bool b)                                      { m_use_default_only_flag = b;    }
  Bool     get_use_default_only_flag   ()                                            { return m_use_default_only_flag; }
  Void     set_pred_mode               (UInt size_idc, UInt list_id, UInt u)         { m_pred_mode[size_idc][list_id] = u;    }
  UInt     get_pred_mode               (UInt size_idc, UInt list_id)                 { return m_pred_mode[size_idc][list_id]; }
  Void     set_pred_qmps_id            (UInt size_idc, UInt list_id, UInt u)         { m_pred_qmps_id[size_idc][list_id] = u;    }
  UInt     get_pred_qmps_id            (UInt size_idc, UInt list_id)                 { return m_pred_qmps_id[size_idc][list_id]; }
  Void     set_pred_size_idc           (UInt size_idc, UInt list_id, UInt u)         { m_pred_size_idc[size_idc][list_id] = u;    }
  UInt     get_pred_size_idc           (UInt size_idc, UInt list_id)                 { return m_pred_size_idc[size_idc][list_id]; }
  Void     set_pred_matrix_id          (UInt size_idc, UInt list_id, UInt u)         { m_pred_matrix_id[size_idc][list_id] = u;    }
  UInt     get_pred_matrix_id          (UInt size_idc, UInt list_id)                 { return m_pred_matrix_id[size_idc][list_id]; }
  Void     set_dpcm_flag               (UInt size_idc, UInt list_id, Bool b)         { m_dpcm_flag[size_idc][list_id] = b;    }
  Bool     get_dpcm_flag               (UInt size_idc, UInt list_id)                 { return m_dpcm_flag[size_idc][list_id]; }
  Void     set_residual_flag           (UInt size_idc, UInt list_id, Bool b)         { m_residual_flag[size_idc][list_id] = b;    }
  Bool     get_residual_flag           (UInt size_idc, UInt list_id)                 { return m_residual_flag[size_idc][list_id]; }
  Void     set_run_flag                (UInt size_idc, UInt list_id, Bool b)         { m_run_flag[size_idc][list_id] = b;    }
  Bool     get_run_flag                (UInt size_idc, UInt list_id)                 { return m_run_flag[size_idc][list_id]; }
  Void     set_Qscale                  (Int i, UInt idx)                             { m_Qscale[idx] = i;          }
#if SONY_QM_AVC_MODE
  Int      get_Qscale                  (UInt idx)                                    { return 1;       }
#else
  Int      get_Qscale                  (UInt idx)                                    { return m_Qscale[idx];       }
#endif
  Void     set_symmetric_flag          (UInt size_idc, UInt list_id, Bool b)         { m_symmetric_flag[size_idc][list_id] = b;  }
  Bool     get_symmetric_flag          (UInt size_idc, UInt list_id)                 { return m_symmetric_flag[size_idc][list_id]; }
  Void     set_vlc_table_num_run       (UInt size_idc, UInt list_id, UInt table_num ){ m_vlc_table_num_run[size_idc][list_id] = table_num;}
  UInt     get_vlc_table_num_run       (UInt size_idc, UInt list_id)                 { return m_vlc_table_num_run[size_idc][list_id];}
  Void     set_vlc_table_num_data      (UInt size_idc, UInt list_id, UInt table_num ){ m_vlc_table_num_data[size_idc][list_id] = table_num;}
  UInt     get_vlc_table_num_data      (UInt size_idc, UInt list_id)                 { return m_vlc_table_num_data[size_idc][list_id];}
  Void     set_sign_flag               (UInt size_idc, UInt list_id, Bool b)         { m_sign_flag[size_idc][list_id] = b;  }
  Bool     get_sign_flag               (UInt size_idc, UInt list_id)                 { return m_sign_flag[size_idc][list_id]; }
  Int*     get_QM_address              (UInt uiQmpsId, UInt uiSizeIdc, UInt uiListId);
  Int*     get_QM_Org_address          (UInt uiQmpsId, UInt uiSizeIdc, UInt uiListId);

  Void     xUpConvMatrix        ( Int* piDst, Int* piSrc, UInt uiDstSizeId, UInt uiSrcSizeId );
  Void     xDownConvMatrix      ( Int* piDst, Int* piSrc, UInt uiDstSizeId, UInt uiSrcSizeId );
  Void     xPredQmatrix         ( TComQM* pcQM_src, Int* piDst, UInt uiDstSizeId, UInt uiDstListId, UInt uiSrcQmpsId, UInt uiSrcSizeIdc, UInt uiSrcMatrixId);
  Void     xQMatrixModeDecision ();
  Void     xCalcBestBit_CopyMode( Int *piOrg, Int * piRecon, Int * piBestRecon, Int iSizeIdc, Int iListIdc, UInt uiSize, UInt *uiBestBit, UInt uiQmpsId );
  Void     xCalcBestBit_DPCMMode( Int *piOrg, Int * piRecon, Int * piBestRecon, Int iSizeIdc, Int iListIdc, UInt uiSize, UInt *uiBestBit, Bool bIsRun );
  UInt     xCalcResidual        (Int *piOrg, Int *piRecon, Int *piResidual, UInt uiSizeIdc, estQMStruct *pestQM);

  UInt     xEstSvlc         ( Int iCode );
  UInt     xEstSvlc         ( Int iCode, UInt uiTableNum );
  UInt     xLeadingZeros    (UInt uiCode);
  UInt     xEstvlc          ( UInt uiCode, UInt num);
  UInt     xConvertToUInt   ( Int iValue ) {  return ( iValue <= 0) ? -iValue<<1 : (iValue<<1)-1; }
  UInt     xEstUvlc         ( UInt uiCode );
  UInt     xEstUvlc         ( UInt uiCode, UInt uiTableNum );
  UInt     xEstFlag         ( Bool b) {return 1; };
  UInt     xPredDPCMQmatrix (Int* piDst, Int* piOrg, UInt uiSizeId, estQMStruct *pestQM);
  Void     xCalcRecon       (Int *piRecon, Int *piResidual, UInt uiSizeIdc);
  Bool     xParseQM         (char* pchFile);
  Bool     IsSymmetry       (Int *piData, UInt uiSizeX, UInt uiErrorTH);
  UInt     xEstResidualCode (UInt uiSize, Int *piData, UInt *puiTableNum, Bool bIsSign);
  UInt     xEstResidualRun  (UInt uiSize, Int *piData, UInt *puiTableNumRun, UInt *puiTableNumData, Bool bIsSign);
  Void     xInvSymmetry     (Int *piDate, UInt uiSizeX);
  Bool     IsSign           (Int *piData, UInt uiSize);
  UInt     xMakeResidual    (Int *piOrg,Int *piRecon, Int *piResidual, UInt uiSizeIdc, Bool bIsQuant);
  Void     xConvZigZag      (Int *piSrc,Int *piDist, UInt uiSizeIdc, Bool bIsSymmetry);
  Void     xMakeDPCM        (Int *piSrc,Int *piDist, UInt uiSizeIdc, Bool bIsSymmetry);
  Void     xMakeDPCMwithQ   (Int* piSrc, Int* piDst, Int* dpcm, UInt uiSizeId, Bool bIsSymmetry);
  Void     xInvZigZag       (Int *piSrc,Int *piDist, UInt uiSizeIdc, Bool bIsSymmetry);
  Void     xInvDPCM         (Int *piSrc,Int *piDist, UInt uiSizeIdc, Bool bIsSymmetry, Int iStartValue);
  Void     xDequantize      (Int *piData, UInt uiSizeIdc, Bool bIsSymmetry);
  Void     set_estQM        (estQMStruct *pestQM, Bool bIsDpcm, Bool bIsRun, UInt uiTableData, UInt uiTableRun, Bool bIsSign);
  Void     xUpdateCondition (UInt iSizeIdc, UInt iListIdc, estQMStruct *pestQM);
#if TI_QM
  Void      set_QMCOption                    ( Int i, UInt idx ) { m_QMCOption[idx]     = i; }
  Int       get_QMCOption                    ( UInt idx )        { return m_QMCOption[idx]; }
#endif
#if SONY_QM_GOP_TEST
  Int      m_iNumFrameCoded;
#endif
};
#endif

/// slice header class
class TComSlice
{
  
private:
  //  Bitstream writing
  Int         m_iPOC;
#if DCM_DECODING_REFRESH
  NalUnitType m_eNalUnitType;         ///< Nal unit type for the slice
#endif
  SliceType   m_eSliceType;
  Int         m_iSliceQp;
  Int         m_iSymbolMode;
  Bool        m_bLoopFilterDisable;
  
  Bool        m_bDRBFlag;             //  flag for future usage as reference buffer
  ERBIndex    m_eERBIndex;            //  flag for future usage as reference buffer
#if DCM_COMB_LIST
  Int         m_aiNumRefIdx   [3];    //  for multiple reference of current slice

  Int         m_iRefIdxOfLC[2][MAX_NUM_REF_LC];
  Int         m_eListIdFromIdxOfLC[MAX_NUM_REF_LC];
  Int         m_iRefIdxFromIdxOfLC[MAX_NUM_REF_LC];
  Int         m_iRefIdxOfL1FromRefIdxOfL0[MAX_NUM_REF_LC];
  Int         m_iRefIdxOfL0FromRefIdxOfL1[MAX_NUM_REF_LC];
  Bool        m_bRefPicListModificationFlagLC;
  Bool        m_bRefPicListCombinationFlag;
#else
  Int         m_aiNumRefIdx   [2];    //  for multiple reference of current slice
#endif  

  //  Data
  Int         m_iSliceQpDelta;
  TComPic*    m_apcRefPicList [2][MAX_NUM_REF];
  Int         m_aiRefPOCList  [2][MAX_NUM_REF];
  Int         m_iDepth;
  
  // referenced slice?
  Bool        m_bRefenced;
#ifdef ROUNDING_CONTROL_BIPRED
  Bool        m_bRounding;
#endif
  
  // access channel
  TComSPS*    m_pcSPS;
  TComPPS*    m_pcPPS;
  TComPic*    m_pcPic;
  
  UInt        m_uiColDir;  // direction to get colocated CUs
  
  Double      m_dLambda;
  
  Bool        m_abEqualRef  [2][MAX_NUM_REF][MAX_NUM_REF];
  
  Bool        m_bNoBackPredFlag;
#if MS_LCEC_LOOKUP_TABLE_EXCEPTION
  Bool        m_bRefIdxCombineCoding;
#endif

  UInt        m_uiSliceMode;
  UInt        m_uiSliceArgument;
  UInt        m_uiSliceCurStartCUAddr;
  UInt        m_uiSliceCurEndCUAddr;
  UInt        m_uiSliceIdx;
  UInt        m_uiEntropySliceMode;
  UInt        m_uiEntropySliceArgument;
  UInt        m_uiEntropySliceCurStartCUAddr;
  UInt        m_uiEntropySliceCurEndCUAddr;
  Bool        m_bNextSlice;
  Bool        m_bNextEntropySlice;
  UInt        m_uiSliceBits;
  
#if SONY_QM
  TComQM*     m_pcQM;                 ///< pointer of quantization matrix
  Bool        m_bQmpsPresentFlag;     ///< flag for using quantization matrix
  UInt        m_uiQmpsId;             ///< QMPS index
#endif
public:
  TComSlice();
  virtual ~TComSlice();
  
  Void      initSlice       ();
  
  Void      setSPS          ( TComSPS* pcSPS ) { m_pcSPS = pcSPS; }
  TComSPS*  getSPS          () { return m_pcSPS; }
  
  Void      setPPS          ( TComPPS* pcPPS ) { m_pcPPS = pcPPS; }
  TComPPS*  getPPS          () { return m_pcPPS; }
  
  SliceType getSliceType    ()                          { return  m_eSliceType;         }
  Int       getPOC          ()                          { return  m_iPOC;           }
  Int       getSliceQp      ()                          { return  m_iSliceQp;           }
  Int       getSliceQpDelta ()                          { return  m_iSliceQpDelta;      }
  Bool      getDRBFlag      ()                          { return  m_bDRBFlag;           }
  ERBIndex  getERBIndex     ()                          { return  m_eERBIndex;          }
  Int       getSymbolMode   ()                          { return  m_iSymbolMode;        }
  Bool      getLoopFilterDisable()                      { return  m_bLoopFilterDisable; }
  Int       getNumRefIdx        ( RefPicList e )                { return  m_aiNumRefIdx[e];             }
  TComPic*  getPic              ()                              { return  m_pcPic;                      }
  TComPic*  getRefPic           ( RefPicList e, Int iRefIdx)    { return  m_apcRefPicList[e][iRefIdx];  }
  Int       getRefPOC           ( RefPicList e, Int iRefIdx)    { return  m_aiRefPOCList[e][iRefIdx];   }
  Int       getDepth            ()                              { return  m_iDepth;                     }
  UInt      getColDir           ()                              { return  m_uiColDir;                   }
  
#if DCM_COMB_LIST 
  Int       getRefIdxOfLC       (RefPicList e, Int iRefIdx)     { return m_iRefIdxOfLC[e][iRefIdx];           }
  Int       getListIdFromIdxOfLC(Int iRefIdx)                   { return m_eListIdFromIdxOfLC[iRefIdx];       }
  Int       getRefIdxFromIdxOfLC(Int iRefIdx)                   { return m_iRefIdxFromIdxOfLC[iRefIdx];       }
  Int       getRefIdxOfL0FromRefIdxOfL1(Int iRefIdx)            { return m_iRefIdxOfL0FromRefIdxOfL1[iRefIdx];}
  Int       getRefIdxOfL1FromRefIdxOfL0(Int iRefIdx)            { return m_iRefIdxOfL1FromRefIdxOfL0[iRefIdx];}
  Bool      getRefPicListModificationFlagLC()                   {return m_bRefPicListModificationFlagLC;}
  Void      setRefPicListModificationFlagLC(Bool bflag)         {m_bRefPicListModificationFlagLC=bflag;}     
  Bool      getRefPicListCombinationFlag()                      {return m_bRefPicListCombinationFlag;}
  Void      setRefPicListCombinationFlag(Bool bflag)            {m_bRefPicListCombinationFlag=bflag;}     
  Void      setListIdFromIdxOfLC(Int  iRefIdx, UInt uiVal)      { m_eListIdFromIdxOfLC[iRefIdx]=uiVal; }
  Void      setRefIdxFromIdxOfLC(Int  iRefIdx, UInt uiVal)      { m_iRefIdxFromIdxOfLC[iRefIdx]=uiVal; }
  Void      setRefIdxOfLC       (RefPicList e, Int iRefIdx, Int RefIdxLC)     { m_iRefIdxOfLC[e][iRefIdx]=RefIdxLC;}
#endif

  Void      setReferenced(Bool b)                               { m_bRefenced = b; }
  Bool      isReferenced()                                      { return m_bRefenced; }
#ifdef ROUNDING_CONTROL_BIPRED
  Void      setRounding(Bool bRound)                            { m_bRounding = bRound; }
  Bool      isRounding()                                        { return m_bRounding; }
#endif
  
  Void      setPOC              ( Int i )                       { m_iPOC              = i;      }
#if DCM_DECODING_REFRESH
  Void      setNalUnitType      ( NalUnitType e )               { m_eNalUnitType      = e;      }
  NalUnitType getNalUnitType    ()                              { return m_eNalUnitType;        }
  Void      decodingRefreshMarking(UInt& uiPOCCDR, Bool& bRefreshPending, TComList<TComPic*>& rcListPic);
#endif
  Void      setSliceType        ( SliceType e )                 { m_eSliceType        = e;      }
  Void      setSliceQp          ( Int i )                       { m_iSliceQp          = i;      }
  Void      setSliceQpDelta     ( Int i )                       { m_iSliceQpDelta     = i;      }
  Void      setDRBFlag          ( Bool b )                      { m_bDRBFlag = b;               }
  Void      setERBIndex         ( ERBIndex e )                  { m_eERBIndex = e;              }
  Void      setSymbolMode       ( Int b  )                      { m_iSymbolMode       = b;      }
  Void      setLoopFilterDisable( Bool b )                      { m_bLoopFilterDisable= b;      }
  
  Void      setRefPic           ( TComPic* p, RefPicList e, Int iRefIdx ) { m_apcRefPicList[e][iRefIdx] = p; }
  Void      setRefPOC           ( Int i, RefPicList e, Int iRefIdx ) { m_aiRefPOCList[e][iRefIdx] = i; }
  Void      setNumRefIdx        ( RefPicList e, Int i )         { m_aiNumRefIdx[e]    = i;      }
  Void      setPic              ( TComPic* p )                  { m_pcPic             = p;      }
  Void      setDepth            ( Int iDepth )                  { m_iDepth            = iDepth; }
  
  Void      setRefPicList       ( TComList<TComPic*>& rcListPic );
  Void      setRefPOCList       ();
  Void      setColDir           ( UInt uiDir ) { m_uiColDir = uiDir; }
  
  Bool      isIntra         ()                          { return  m_eSliceType == I_SLICE;  }
  Bool      isInterB        ()                          { return  m_eSliceType == B_SLICE;  }
  Bool      isInterP        ()                          { return  m_eSliceType == P_SLICE;  }
  
  Void      setLambda( Double d ) { m_dLambda = d; }
  Double    getLambda() { return m_dLambda;        }
  
  Void      initEqualRef();
  Bool      isEqualRef  ( RefPicList e, Int iRefIdx1, Int iRefIdx2 )
  {
    if (iRefIdx1 < 0 || iRefIdx2 < 0) return false;
    return m_abEqualRef[e][iRefIdx1][iRefIdx2];
  }
  
  Void setEqualRef( RefPicList e, Int iRefIdx1, Int iRefIdx2, Bool b)
  {
    m_abEqualRef[e][iRefIdx1][iRefIdx2] = m_abEqualRef[e][iRefIdx2][iRefIdx1] = b;
  }
  
  static Void      sortPicList         ( TComList<TComPic*>& rcListPic );
  
  Bool getNoBackPredFlag() { return m_bNoBackPredFlag; }
  Void setNoBackPredFlag( Bool b ) { m_bNoBackPredFlag = b; }
#if MS_LCEC_LOOKUP_TABLE_EXCEPTION
  Bool getRefIdxCombineCoding() { return m_bRefIdxCombineCoding; }
  Void setRefIdxCombineCoding( Bool b ) { m_bRefIdxCombineCoding = b; }
#endif
#if DCM_COMB_LIST
  Void      generateCombinedList       ();
#endif
  Void setSliceMode                     ( UInt uiMode )     { m_uiSliceMode = uiMode;                     }
  UInt getSliceMode                     ()                  { return m_uiSliceMode;                       }
  Void setSliceArgument                 ( UInt uiArgument ) { m_uiSliceArgument = uiArgument;             }
  UInt getSliceArgument                 ()                  { return m_uiSliceArgument;                   }
  Void setSliceCurStartCUAddr           ( UInt uiAddr )     { m_uiSliceCurStartCUAddr = uiAddr;           }
  UInt getSliceCurStartCUAddr           ()                  { return m_uiSliceCurStartCUAddr;             }
  Void setSliceCurEndCUAddr             ( UInt uiAddr )     { m_uiSliceCurEndCUAddr = uiAddr;             }
  UInt getSliceCurEndCUAddr             ()                  { return m_uiSliceCurEndCUAddr;               }
  Void setSliceIdx                      ( UInt i)           { m_uiSliceIdx = i;                           }
  UInt getSliceIdx                      ()                  { return  m_uiSliceIdx;                       }
  Void copySliceInfo                    (TComSlice *pcSliceSrc);
  Void setEntropySliceMode              ( UInt uiMode )     { m_uiEntropySliceMode = uiMode;              }
  UInt getEntropySliceMode              ()                  { return m_uiEntropySliceMode;                }
  Void setEntropySliceArgument          ( UInt uiArgument ) { m_uiEntropySliceArgument = uiArgument;      }
  UInt getEntropySliceArgument          ()                  { return m_uiEntropySliceArgument;            }
  Void setEntropySliceCurStartCUAddr    ( UInt uiAddr )     { m_uiEntropySliceCurStartCUAddr = uiAddr;    }
  UInt getEntropySliceCurStartCUAddr    ()                  { return m_uiEntropySliceCurStartCUAddr;      }
  Void setEntropySliceCurEndCUAddr      ( UInt uiAddr )     { m_uiEntropySliceCurEndCUAddr = uiAddr;      }
  UInt getEntropySliceCurEndCUAddr      ()                  { return m_uiEntropySliceCurEndCUAddr;        }
  Void setNextSlice                     ( Bool b )          { m_bNextSlice = b;                           }
  Bool isNextSlice                      ()                  { return m_bNextSlice;                        }
  Void setNextEntropySlice              ( Bool b )          { m_bNextEntropySlice = b;                    }
  Bool isNextEntropySlice               ()                  { return m_bNextEntropySlice;                 }
  Void setSliceBits                     ( UInt uiVal )      { m_uiSliceBits = uiVal;                      }
  UInt getSliceBits                     ()                  { return m_uiSliceBits;                       }  
#if SONY_QM
  Void      setQM              ( TComQM* pcQM ) { m_pcQM = pcQM; }
  TComQM*   getQM              ()               { return m_pcQM; }
  Void      setDefaultQM       ();
  Void      setQmpsID          (UInt ui       ) { m_uiQmpsId = ui; }
  UInt      getQmpsID          ()               { return m_uiQmpsId; }
  Void      setQmpsPresentFlag (Bool b        ) { m_bQmpsPresentFlag = b; }
  Bool      getQmpsPresentFlag ()               { return m_bQmpsPresentFlag; }
  Bool      CheckDefaultQM     ();
#endif

protected:
  TComPic*  xGetRefPic  (TComList<TComPic*>& rcListPic,
                         Bool                bDRBFlag,
                         ERBIndex            eERBIndex,
                         UInt                uiPOCCurr,
                         RefPicList          eRefPicList,
                         UInt                uiNthRefPic );
  
};// END CLASS DEFINITION TComSlice


#endif // __TCOMSLICE__

